class ApplicationFilter(object):
    ExpandScores = 'scores'
    ExpandTraceBreakdown = 'trace_breakdown'
    ExpandLicense = 'license'
    ExpandTechnologies = 'technologies'

    def __init__(self):
        self.expand = []
        self.filter_text = None
        self.filter_servers = []
        self.filter_techs = []
        self.filter_tags = []
        self.filter_languages = []
        self.include_archived = False
        self.include_merged = True
        self.limit = 20
        self.offset = 0
        self.sort = '-appName'


    def get_params_as_json(self):
        return {
            'expand': ','.join(self.expand),
            'filterText': self.filter_text,
            'filterServers': ','.join(self.filter_servers),
            'filterTechs': ','.join(self.filter_techs),
            'filterTags': ','.join(self.filter_tags),
            'filterLanguages': ','.join(self.filter_languages),
            'includeArchived': self.include_archived,
            'includeMerged': self.include_merged,
            'limit': self.limit,
            'offset': self.offset,
            'sort': self.sort
        }


from .api_support import _ApiSupport
from ..filters.trace_filter import TraceFilter
from ..filters.trace_trend_filter import TraceTrendFilter
from ..filters.trace_time_to_remediate_filter import TraceTimeToRemediateFilter


class _TraceApi(_ApiSupport):

    def __init__(self):
        super(_TraceApi, self).__init__()

    def filter_org_traces(self, org_uuid, trace_filter=None):
        if trace_filter is None:
            trace_filter = TraceFilter()
        path = '{org_uuid}/orgtraces/filter'.format(org_uuid=org_uuid)
        return self._get(path, params=trace_filter.get_params_as_json())

    def get_org_trace(self, org_uuid, trace_id, expand=None):
        path = '{org_uuid}/orgtraces/filter/{trace_id}'.format(
            org_uuid=org_uuid,
            trace_id=trace_id
        )
        return self._get(path, params={'expand': expand})

    def get_trace_notes(self, org_uuid, app_id, trace_id):
        path = '{org_uuid}/applications/{app_id}/traces/{trace_id}/notes'.format(
            org_uuid=org_uuid,
            app_id=app_id,
            trace_id=trace_id
        )
        return self._get(path)

    def create_trace_note(self,org_uuid, app_id, trace_id, note):
        path = '{org_uuid}/applications/{app_id}/traces/{trace_id}/notes'.format(
            org_uuid=org_uuid,
            app_id=app_id,
            trace_id=trace_id
        )
        return self._post(path, data={'note': note})

    def get_org_trace_policy_violations(self, org_uuid):
        path = '{org_uuid}/orgtraces/policy/violations'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_trace_visibility(self, org_uuid, trace_id):
        path = '{org_uuid}/orgtraces/{trace_id}/visible'.format(
            org_uuid=org_uuid,
            trace_id=trace_id
        )
        return self._get(path)

    def get_new_trace_trend(self, org_uuid, interval='week', trace_trend_filter=None):
        if trace_trend_filter is None:
            trace_trend_filter = TraceTrendFilter()
        path = '{org_uuid}/orgtraces/stats/trend/{interval}/new'.format(
            org_uuid=org_uuid,
            interval=interval
        )
        return self._get(path, params=trace_trend_filter.get_params_as_json())

    def get_total_trace_trend(self, org_uuid, interval='week', trace_trend_filter=None):
        if trace_trend_filter is None:
            trace_trend_filter = TraceTrendFilter()
        path = '{org_uuid}/orgtraces/stats/trend/{interval}/total'.format(
            org_uuid=org_uuid,
            interval=interval
        )
        return self._get(path, params=trace_trend_filter.get_params_as_json())


    def get_trace_time_to_remediate_month_trend(self, org_uuid):
        path = '{org_uuid}/orgtraces/stats/ttr/severity/trend'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_trace_card(self, org_uuid, trace_uuid):
        path = '{org_uuid}/traces/{trace_uuid}/card'.format(
            org_uuid=org_uuid,
            trace_uuid=trace_uuid
        )
        return self._get(path)

    def get_trace_events_summary(self, org_uuid, trace_uuid):
        path = '{org_uuid}/traces/{trace_uuid}/events/summary'.format(
            org_uuid=org_uuid,
            trace_uuid=trace_uuid
        )
        return self._get(path)

    def get_trace_event_details(self, org_uuid, trace_uuid, event_id):
        path = '{org_uuid}/traces/{trace_uuid}/events/{event_id}/details'.format(
            org_uuid=org_uuid,
            trace_uuid=trace_uuid,
            event_id=event_id
        )
        return self._get(path)

    def get_trace_httprequest(self, org_uuid, trace_uuid):
        path = '{org_uuid}/traces/{trace_uuid}/httprequest'.format(
            org_uuid=org_uuid,
            trace_uuid=trace_uuid,
        )
        return self._get(path)

    def get_trace_httprequest_details(self, org_uuid, trace_uuid):
        path = '{org_uuid}/traces/{trace_uuid}/httprequest/details'.format(
            org_uuid=org_uuid,
            trace_uuid=trace_uuid,
        )
        return self._get(path)

    def get_trace_recommendation(self, org_uuid, trace_uuid):
        path = '{org_uuid}/traces/{trace_uuid}/recommendation'.format(
            org_uuid=org_uuid,
            trace_uuid=trace_uuid,
        )
        return self._get(path)

    def get_trace_story(self, org_uuid, trace_uuid):
        path = '{org_uuid}/traces/{trace_uuid}/story'.format(
            org_uuid=org_uuid,
            trace_uuid=trace_uuid,
        )
        return self._get(path)



import base64
import validators


class Util:

    def __init__(self):
        pass

    @staticmethod
    def create_authorization_token(username, service_key):
        data_string = "{}:{}".format(username,service_key)
        data_bytes = data_string.encode("UTF-8")
        result = base64.b64encode(data_bytes)
        return result.decode('ascii')

    @staticmethod
    def validate_url(url):
        return validators.url(url)



from .api_support import _ApiSupport


class _AlertApi(_ApiSupport):

    def __init__(self):
        super(_AlertApi, self).__init__()

    def get_alerts(self, org_uuid):
        path = '{org_uuid}/alerts'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_alert_data(self, org_uuid, alert_id):
        path = '{org_uuid}/alerts/{alert_id}'.format(org_uuid=org_uuid, alert_id=alert_id)
        return self._get(path)


from .api_support import _ApiSupport


class _ScoresApi(_ApiSupport):

    def __init__(self):
        super(_ScoresApi, self).__init__()

    def get_overall_scores(self, org_uuid):
        path = '{org_uuid}/scores'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_score_server_breakdown(self, org_uuid):
        path = '{org_uuid}/scores/breakdown/server'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_score_trace_rule_breakdown(self, org_uuid):
        path = '{org_uuid}/scores/breakdown/trace/rule'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_score_trace_severity_breakdown(self, org_uuid):
        path = '{org_uuid}/scores/breakdown/trace/severity'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_score_trace_status_breakdown(self, org_uuid):
        path = '{org_uuid}/scores/breakdown/trace/status'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_score_platform(self, org_uuid):
        path = '{org_uuid}/scores/platform'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_score_platform_include_defense(self, org_uuid):
        path = '{org_uuid}/scores/platform/defense'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_score_security(self, org_uuid):
        path = '{org_uuid}/scores/security'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_score_security_include_defense(self, org_uuid):
        path = '{org_uuid}/scores/security/defense'.format(org_uuid=org_uuid)
        return self._get(path)


from .api_support import _ApiSupport
from ..filters.application_library_filter import ApplicationLibraryFilter
from ..filters.application_filter import ApplicationFilter
from ..filters.application_trace_filter import ApplicationTraceFilter

class _ApplicationApi(_ApiSupport):

    def __init__(self):
        super(_ApplicationApi, self).__init__()

    # App Activity
    def get_inactive_applications(self, org_uuid, expand=None, include_archived=False, include_merged=False, limit=5):
        path = '{org_uuid}/applications/activity/inactive'.format(org_uuid=org_uuid)
        return self._get(path, params={'expand': expand, 'includeArchived': include_archived, 'includeMerged': include_merged, 'limit': limit})

    def get_newest_applications(self, org_uuid, expand=None, include_archived=False, include_merged=False, limit=5):
        path = '{org_uuid}/applications/activity/newest'.format(org_uuid=org_uuid)
        return self._get(path, params={'expand': expand, 'includeArchived': include_archived, 'includeMerged': include_merged, 'limit': limit})

    def get_recent_applications(self, org_uuid, expand=None, include_archived=False, include_merged=False, limit=5):
        path = '{org_uuid}/applications/activity/recent'.format(org_uuid=org_uuid)
        return self._get(path, params={'expand': expand, 'includeArchived': include_archived, 'includeMerged': include_merged, 'limit': limit})

    # App Agent Activity
    def get_application_agent_activity(self, org_uuid, app_id, range, include_merged=False):
        path = '{org_uuid}/applications/{app_id}/agent/activity/{range}'.format(org_uuid=org_uuid, app_id=app_id, range=range)
        return self._get(path, params={'includeMerged': include_merged})

    # App Components
    def get_application_components(self,org_uuid, app_id):
        path = '{org_uuid}/applications/{app_id}/components'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path)

    # App Coverage
    def get_application_coverage(self,org_uuid, app_id, include_merged=True, limit=50):
        path = '{org_uuid}/applications/{app_id}/coverage'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'includeMerged': include_merged, 'limit': limit})

    def get_application_coverage_past_week(self,org_uuid, app_id, include_merged=True, limit=50):
        path = '{org_uuid}/applications/{app_id}/coverage/stats/week'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'includeMerged': include_merged, 'limit': limit})

    # App History
    def get_application_history(self, org_uuid, app_id, include_merged=True):
        path = '{org_uuid}/applications/{app_id}/history'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'includeMerged': include_merged})

    def get_application_history_by_interval(self, org_uuid, app_id, environment='DEVELOPMENT', interval='WEEK', include_merged=True, include_defense=False):
        defense_string = '/defense' if include_defense else ''
        path = '{org_uuid}/applications/{app_id}/history/interval{defense_string}'.format(org_uuid=org_uuid, app_id=app_id, defense_string=defense_string)
        return self._get(path, params={'environment': environment, 'interval': interval, 'includeMerged': include_merged})

    # App libraries
    def get_application_libraries(self, org_uuid, app_id, expand=None, load_cve=False, quick_filter='ALL'):
        path = '{org_uuid}/applications/{app_id}/libraries'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'expand': expand, 'loadCVE': load_cve, 'quickFilter': quick_filter})

    def filter_application_libraries(self, org_uuid, app_id, app_library_filter=None):
        if app_library_filter is None:
            app_library_filter = ApplicationLibraryFilter()
        path = '{org_uuid}/applications/{app_id}/libraries/filter'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params=app_library_filter.get_params_as_json())

    def get_application_library_subfilters(self, org_uuid, app_id, libraries_filter_type):
        path = '{org_uuid}/applications/{app_id}/libraries/filters/{libraries_filter_type}/listing'.format(org_uuid=org_uuid, app_id=app_id, libraries_filter_type=libraries_filter_type)
        return self._get(path)

    def get_application_libraries_stats(self, org_uuid, app_id, include_merged=False):
        path = '{org_uuid}/applications/{app_id}/libraries/stats'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'includeMerged': include_merged})

    # App Scores

    def get_application_trace_breakdown(self, org_uuid, app_id, include_merged=True):
        path = '{org_uuid}/applications/{app_id}/breakdown/trace'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'includeMerged': include_merged})

    def get_application_trace_rule_breakdown(self, org_uuid, app_id, environment='DEVELOPMENT'):
        path = '{org_uuid}/applications/{app_id}/breakdown/trace/rule'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'environment': environment})

    def get_application_trace_status_breakdown(self, org_uuid, app_id, environment='DEVELOPMENT'):
        path = '{org_uuid}/applications/{app_id}/breakdown/trace/status'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'environment': environment})

    # App Servers
    def get_application_servers(self, org_uuid, app_id, expand=None, include_merged=True, only_licensed=False):
        path = '{org_uuid}/applications/{app_id}/servers'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'expand': expand, 'includeMerged': include_merged, 'onlyLicensed': only_licensed})

    def get_application_servers_breakdown(self, org_uuid, app_id):
        path = '{org_uuid}/applications/{app_id}/servers/breakdown'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path)

    def get_application_servers_count(self, org_uuid, app_id, include_merged=True):
        path = '{org_uuid}/applications/{app_id}/servers/count'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'includeMerged': include_merged})

    def get_application_servers_recently_active(self, org_uuid, app_id, expand=None, include_merged=True):
        path = '{org_uuid}/applications/{app_id}/servers/newest'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'expand': expand, 'includeMerged': include_merged})

    def get_application_servers_properties(self, org_uuid, app_id, include_merged=True):
        path = '{org_uuid}/applications/{app_id}/servers/properties'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'includeMerged': include_merged})

    def get_application_servers_settings(self, org_uuid, app_id, include_merged=True, filter_environment=None):
        env_string = '/environment' if filter_environment is not None else ''
        path = '{org_uuid}/applications/{app_id}/servers/settings{env_string}'.format(org_uuid=org_uuid, app_id=app_id, env_string=env_string)
        return self._get(path, params={'includeMerged': include_merged, 'environment': filter_environment})

    # App Technologies
    def get_application_technologies(self, org_uuid, app_id):
        path = '{org_uuid}/applications/{app_id}/techs'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path)

    # Techs
    def get_technologies(self, org_uuid):
        path = '{org_uuid}/techs'.format(org_uuid=org_uuid)
        return self._get(path)

    # Applications
    def get_total_allowed_applications(self, org_uuid):
        path = '{org_uuid}/applications/allowed'.format(org_uuid=org_uuid)
        return self._get(path)

    def filter_applications(self, org_uuid, application_filter=None):
        if application_filter is None:
            application_filter = ApplicationFilter()
        path = '{org_uuid}/applications/filter'.format(org_uuid=org_uuid)
        return self._post(path, params=application_filter.get_params_as_json())

    def get_application_filters(self, org_uuid):
        path = '{org_uuid}/applications/filters/listing'.format(org_uuid=org_uuid)
        return self._post(path)

    def get_application(self, org_uuid, app_id, expand=None, include_merged=True):
        path = '{org_uuid}/applications/{app_id}'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'expand': expand, 'includeMerged': include_merged})

    def update_application_importance(self, org_uuid, app_id, importance):
        path = '{org_uuid}/applications/{app_id}/importance'.format(org_uuid=org_uuid, app_id=app_id)
        return self._put(path, data={'importance': importance})

    def get_application_license_details(self, org_uuid, app_id):
        path = '{org_uuid}/applications/{app_id}/license'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path)

    # Application Trace Filtering
    def filter_application_traces(self, org_uuid, app_id, application_trace_filter=None):
        if application_trace_filter is None:
            application_trace_filter = ApplicationTraceFilter()
        path = '{org_uuid}/traces/{app_id}/filter'.format(org_uuid=org_uuid, app_id=app_id)
        return self._post(path, params=application_trace_filter.get_params_as_json())

    def get_application_vuln_details(self, org_uuid, app_id, trace_uuid, expand=None):
        path = '{org_uuid}/traces/{app_id}/filter/{trace_uuid}'.format(org_uuid=org_uuid, app_id=app_id, trace_uuid=trace_uuid)
        return self._get(path, params={'expand': expand})

    def get_application_traces_with_policy_violations(self, org_uuid, app_id, environment='DEVELOPMENT'):
        path = '{org_uuid}/traces/{app_id}/policy/violations'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'environment': environment})

    def delete_application_trace(self, org_uuid, app_id, trace_id):
        path = '{org_uuid}/traces/{app_id}/trace/{trace_id}'.format(org_uuid=org_uuid, app_id=app_id, trace_id=trace_id)
        return self._delete(path)

    def delete_application_traces(self,org_uuid, app_id, trace_id_array):
        path = '{org_uuid}/traces/{app_id}'.format(org_uuid=org_uuid, app_id=app_id)
        return self._delete(path, data={'traces': trace_id_array})

    def get_application_trace_details(self, org_uuid, app_id, trace_id, expand=None):
        path = '{org_uuid}/traces/{app_id}/trace/{trace_id}'.format(org_uuid=org_uuid, app_id=app_id, trace_id=trace_id)
        return self._get(path, params={'expand': expand})

    def get_application_trace_requirements(self, org_uuid, app_id, trace_id, expand=None):
        path = '{org_uuid}/traces/{app_id}/trace/{trace_id}/requirements'.format(org_uuid=org_uuid, app_id=app_id, trace_id=trace_id)
        return self._get(path, params={'expand': expand})

    def get_application_trace_servers(self, org_uuid, app_id, trace_id, expand=None):
        path = '{org_uuid}/traces/{app_id}/trace/{trace_id}/servers'.format(org_uuid=org_uuid, app_id=app_id, trace_id=trace_id)
        return self._get(path, params={'expand': expand})

    def get_application_trace_urls(self, org_uuid, app_id, trace_id, expand=None):
        path = '{org_uuid}/traces/{app_id}/trace/{trace_id}/urlinstances'.format(org_uuid=org_uuid, app_id=app_id, trace_id=trace_id)
        return self._get(path, params={'expand': expand})

    def get_application_trace_visibility(self, org_uuid, app_id, trace_id):
        path = '{org_uuid}/traces/{app_id}/{trace_id}/visible'.format(org_uuid=org_uuid, app_id=app_id, trace_id=trace_id)
        return self._get(path)








class TraceTrendFilter(object):

    def __init__(self):
        self.rule_severities = []
        self.rules = []
        self.all_rules = None
        self.application_importance = None
        self.application_ids = []
        self.all_applications = None
        self.server_environments = []
        self.servers = []
        self.all_servers = None

    def get_params_as_json(self):
        return {
            'ruleSeverities':  ','.join(self.rule_severities),
            'rules': ','.join(self.rules),
            'allRules': self.all_rules,
            'applicationImportance': self.application_importance,
            'applicationIds': self.application_ids,
            'allApplications': self.all_applications,
            'serverEnvironments': self.server_environments,
            'servers': self.servers,
            'allServers': self.all_servers
        }


from .user_api.util import Util
from .user_api.organization_api import _OrganizationApi
from .user_api.application_api import _ApplicationApi
from .user_api.server_api import _ServerApi
from .user_api.tags_api import _TagsApi
from .user_api.alerts_api import _AlertApi
from .user_api.modules_api import _ModulesApi
from .user_api.library_api import _LibraryApi
from .user_api.scores_api import _ScoresApi
from .user_api.history_api import _HistoryApi
from .user_api.role_api import _RoleApi
from .user_api.profile_api import _ProfileApi
from .user_api.policy_api import _PolicyApi
from .user_api.webhook_api import _WebhookApi
from .user_api.user_api import _UserApi
from .user_api.trace_api import _TraceApi
from .user_api.route_coverage import _RouteCoverageApi
from .user_api.session_metadata_api import _SessionMetadataApi

class ContrastSdk(object):

    def __init__(self, username, api_key, service_key, teamserver_url='https://app.contrastsecurity.com'):
        self._username = username
        self._api_key = api_key
        self._service_key = service_key
        self._teamserver_url = teamserver_url

        self._setup_apis()

    def _configure_api_defaults(self, api_class):
        api_class._headers = self._create_headers()
        api_class._base_url = self._teamserver_url + '/Contrast/api'

    def _create_headers(self):
        return {
            'Authorization': Util.create_authorization_token(self._username, self._service_key),
            'API-Key': self._api_key,
            'Content-type': 'application/json',
            'Accept': 'application/json'
        }

    def _setup_apis(self):
        self._configure_organization_api()
        self._configure_server_api()
        self._configure_tags_api()
        self._configure_alert_api()
        self._configure_modules_api()
        self._configure_library_api()
        self._configure_scores_api()
        self._configure_history_api()
        self._configure_roles_api()
        self._configure_profile_api()
        self._configure_application_api()
        self._configure_policy_api()
        self._configure_user_api()
        self._configure_trace_api()
        self._configure_webhook_api()
        self._configure_route_coverage_api()
        self._configure_session_metadata_api()

    def _configure_application_api(self):
        self._applications = _ApplicationApi()
        self._configure_api_defaults(self._applications)
        self.get_inactive_applications = self._applications.get_inactive_applications
        self.get_newest_applications = self._applications.get_newest_applications
        self.get_recent_applications = self._applications.get_recent_applications
        self.get_application_agent_activity = self._applications.get_application_agent_activity
        self.get_application_components = self._applications.get_application_components
        self.get_application_coverage = self._applications.get_application_coverage
        self.get_application_coverage_past_week = self._applications.get_application_coverage_past_week
        self.get_application_history = self._applications.get_application_history
        self.get_application_history_by_interval = self._applications.get_application_history_by_interval
        self.get_application_libraries = self._applications.get_application_libraries
        self.filter_application_libraries = self._applications.filter_application_libraries
        self.get_application_library_subfilters = self._applications.get_application_library_subfilters
        self.get_application_libraries_stats = self._applications.get_application_libraries_stats
        self.get_application_trace_breakdown = self._applications.get_application_trace_breakdown
        self.get_application_trace_rule_breakdown = self._applications.get_application_trace_rule_breakdown
        self.get_application_trace_status_breakdown = self._applications.get_application_trace_status_breakdown
        self.get_application_servers = self._applications.get_application_servers
        self.get_application_servers_breakdown = self._applications.get_application_servers_breakdown
        self.get_application_servers_count = self._applications.get_application_servers_count
        self.get_application_servers_recently_active = self._applications.get_application_servers_recently_active
        self.get_application_servers_properties = self._applications.get_application_servers_properties
        self.get_application_servers_settings = self._applications.get_application_servers_settings
        self.get_application_technologies = self._applications.get_application_technologies
        self.get_technologies = self._applications.get_technologies
        self.get_total_allowed_applications = self._applications.get_total_allowed_applications
        self.filter_applications = self._applications.filter_applications
        self.get_application_filters = self._applications.get_application_filters
        self.get_application = self._applications.get_application
        self.update_application_importance = self._applications.update_application_importance
        self.get_application_license_details = self._applications.get_application_license_details
        self.filter_application_traces = self._applications.filter_application_traces
        self.get_application_vuln_details = self._applications.get_application_vuln_details
        self.get_application_traces_with_policy_violations = self._applications.get_application_traces_with_policy_violations
        self.delete_application_trace = self._applications.delete_application_trace
        self.delete_application_traces = self._applications.delete_application_traces
        self.get_application_trace_details = self._applications.get_application_trace_details
        self.get_application_trace_requirements = self._applications.get_application_trace_requirements
        self.get_application_trace_servers = self._applications.get_application_trace_servers
        self.get_application_trace_urls = self._applications.get_application_trace_urls
        self.get_application_trace_visibility = self._applications.get_application_trace_visibility

    def _configure_trace_api(self):
        self._traces = _TraceApi()
        self._configure_api_defaults(self._traces)
        self.filter_org_traces = self._traces.filter_org_traces
        self.get_org_trace = self._traces.get_org_trace
        self.get_trace_notes = self._traces.get_trace_notes
        self.create_trace_note = self._traces.create_trace_note
        self.get_org_trace_policy_violations = self._traces.get_org_trace_policy_violations
        self.get_trace_visibility = self._traces.get_trace_visibility
        self.get_new_trace_trend = self._traces.get_new_trace_trend
        self.get_total_trace_trend = self._traces.get_total_trace_trend
        self.get_trace_time_to_remediate_month_trend = self._traces.get_trace_time_to_remediate_month_trend
        self.get_trace_card = self._traces.get_trace_card
        self.get_trace_events_summary = self._traces.get_trace_events_summary
        self.get_trace_event_details = self._traces.get_trace_event_details
        self.get_trace_httprequest = self._traces.get_trace_httprequest
        self.get_trace_httprequest_details = self._traces.get_trace_httprequest_details
        self.get_trace_recommendation = self._traces.get_trace_recommendation
        self.get_trace_story = self._traces.get_trace_story

    def _configure_alert_api(self):
        self._alert = _AlertApi()
        self._configure_api_defaults(self._alert)
        self.get_alerts = self._alert.get_alerts
        self.get_alert_data = self._alert.get_alert_data


    def _configure_tags_api(self):
        self._tags = _TagsApi()
        self._configure_api_defaults(self._tags)
        self.get_application_tags = self._tags.get_application_tags
        self.get_all_library_tags = self._tags.get_all_library_tags
        self.get_all_application_tags = self._tags.get_all_application_tags
        self.get_application_library_tags = self._tags.get_application_library_tags
        self.get_library_tag_list = self._tags.get_library_tag_list
        self.get_server_tag_list = self._tags.get_server_tag_list
        self.get_all_server_tags = self._tags.get_all_server_tags
        self.get_all_trace_tags = self._tags.get_all_trace_tags
        self.get_all_trace_tags_for_application = self._tags.get_all_trace_tags_for_application
        self.get_all_trace_tags_for_servers = self._tags.get_all_trace_tags_for_servers
        self.get_all_tags_for_trace = self._tags.get_all_tags_for_trace
        self.tag_application = self._tags.tag_application
        self.tag_library = self._tags.tag_library
        self.tag_server = self._tags.tag_server
        self.tag_trace = self._tags.tag_trace
        self.delete_tag_from_application = self._tags.delete_tag_from_application
        self.delete_tag_from_trace = self._tags.delete_tag_from_trace
        self.delete_tag_from_server = self._tags.delete_tag_from_server
        self.delete_tag_from_library = self._tags.delete_tag_from_library

    def _configure_server_api(self):
        self._server = _ServerApi()
        self._configure_api_defaults(self._server)
        self.get_servers = self._server.get_servers
        self.get_active_servers = self._server.get_active_servers
        self.filter_servers = self._server.filter_servers
        self.get_server_filters = self._server.get_server_filters
        self.get_server_filter_subfilters = self._server.get_server_filter_subfilters
        self.get_server_modes = self._server.get_server_modes
        self.get_server_details = self._server.get_server_details
        self.get_server_activity = self._server.get_server_activity
        self.get_server_agent_activity = self._server.get_server_agent_activity
        self.get_server_attack_status = self._server.get_server_attack_status
        self.get_server_attack_types = self._server.get_server_attack_types
        self.get_server_trace_breakdown = self._server.get_server_trace_breakdown
        self.get_server_trace_severity_breakdown = self._server.get_server_trace_severity_breakdown
        self.get_server_trace_status_breakdwon = self._server.get_server_trace_status_breakdown
        self.get_server_libraries_breakdown = self._server.get_server_libraries_breakdown
        self.update_server_name = self._server.update_server_name
        self.get_server_properties = self._server.get_server_properties
        self.get_server_vuln_and_attack_urls = self._server.get_server_vuln_and_attack_urls
        self.get_server_vuln_urls = self._server.get_server_vuln_urls
        self.get_server_attack_urls = self._server.get_server_attack_urls
        self.get_server_libraries = self._server.get_server_libraries
        self.get_server_libraries_subfilters = self._server.get_server_libraries_subfilters
        self.filter_server_libraries = self._server.filter_server_libraries
        self.get_server_trace_subfilters = self._server.get_server_trace_subfilters
        self.get_server_trace_details = self._server.get_server_trace_details
        self.filter_server_traces = self._server.filter_server_traces
        self.delete_server_traces = self._server.delete_server_traces
        self.get_server_policy_violations = self._server.get_server_policy_violations
        self.delete_server_trace = self._server.delete_server_trace
        self.get_server_trace_vulnerability = self._server.get_server_trace_vulnerability

    def _configure_modules_api(self):
        self._modules = _ModulesApi()
        self._configure_api_defaults(self._modules)
        self.get_application_modules = self._modules.get_application_modules
        self.get_application_child_modules = self._modules.get_application_child_modules

    def _configure_history_api(self):
        self._history = _HistoryApi()
        self._configure_api_defaults(self._history)
        self.get_organization_score_history = self._history.get_organization_score_history
        self.get_organization_score_history_interval = self._history.get_organization_score_history_interval

    def _configure_roles_api(self):
        self._roles = _RoleApi()
        self._configure_api_defaults(self._roles)
        self.get_roles = self._roles.get_roles

    def _configure_scores_api(self):
        self._scores = _ScoresApi()
        self._configure_api_defaults(self._scores)
        self.get_overall_scores = self._scores.get_overall_scores
        self.get_score_server_breakdown = self._scores.get_score_server_breakdown
        self.get_score_trace_rule_breakdown = self._scores.get_score_trace_rule_breakdown
        self.get_score_trace_severity_breakdown = self._scores.get_score_trace_severity_breakdown
        self.get_score_trace_status_breakdown = self._scores.get_score_trace_status_breakdown
        self.get_score_platform = self._scores.get_score_platform
        self.get_score_platform_include_defense = self._scores.get_score_platform_include_defense
        self.get_score_security = self._scores.get_score_security
        self.get_score_security_include_defense = self._scores.get_score_security_include_defense

    def _configure_library_api(self):
        self._library = _LibraryApi()
        self._configure_api_defaults(self._library)
        self.get_libraries = self._library.get_libraries
        self.get_dotnet_library = self._library.get_dotnet_library
        self.get_java_library = self._library.get_java_library
        self.get_library_stats = self._library.get_library_stats
        self.get_library_filter_subfilters = self._library.get_library_filter_subfilters
        self.filter_libraries = self._library.filter_libraries
        self.get_all_library_filters = self._library.get_all_library_filters
        self.get_library_policy = self._library.get_library_policy

    def _configure_profile_api(self):
        self._profile = _ProfileApi()
        self._configure_api_defaults(self._profile)
        self.get_profile_info = self._profile.get_profile_info
        self.get_profile_organizations = self._profile.get_profile_organizations
        self.get_profile_default_organization = self._profile.get_profile_default_organization
        self.get_org_info = self._profile.get_org_info
        self.get_profile_password_policy = self._profile.get_profile_password_policy
        self.get_profile_roles = self._profile.get_profile_roles
        self.set_profile_default_org = self._profile.set_profile_default_org

    def _configure_organization_api(self):
        self._organization = _OrganizationApi()
        self._configure_api_defaults(self._organization)
        self.search = self._organization.search
        self.get_organization_info = self._organization.get_organization_info
        self.get_organization_administrators = self._organization.get_organization_administrators
        self.get_organization_application_roles = self._organization.get_organization_application_roles
        self.get_organization_library_scoring = self._organization.get_organization_library_scoring
        self.put_organization_library_scoring = self._organization.put_organization_library_scoring
        self.get_organization_servers_needing_restart = self._organization.get_organization_servers_needing_restart
        self.get_organization_application_stats = self._organization.get_organization_application_stats
        self.get_organization_server_stats = self._organization.get_organization_server_stats
        self.get_organization_library_stats = self._organization.get_organization_library_stats
        self.get_organization_trace_stats = self._organization.get_organization_trace_stats
        self.get_organization_server_settings = self._organization.get_organization_server_settings

    def _configure_policy_api(self):
        self._policy = _PolicyApi()
        self._configure_api_defaults(self._policy)
        self.get_validators_and_sanitizers = self._policy.get_validators_and_sanitizers
        self.get_validator_controls = self._policy.get_validator_controls
        self.get_sanitizer_controls = self._policy.get_sanitizer_controls
        self.get_control_suggestions = self._policy.get_control_suggestions

    def _configure_webhook_api(self):
        self._webhook = _WebhookApi()
        self._configure_api_defaults(self._webhook)
        self.get_webhooks = self._webhook.get_webhooks
        self.get_webhook = self._webhook.get_webhook

    def _configure_user_api(self):
        self._user = _UserApi()
        self._configure_api_defaults(self._user)
        self.get_users = self._user.get_users
        self.get_custom_alerts = self._user.get_custom_alerts
        self.get_custom_attack_alerts = self._user.get_custom_attack_alerts
        self.get_custom_vulnerability_alerts = self._user.get_custom_vulnerability_alerts
        self.get_user_information = self._user.get_user_information
        self.get_user_authorization_header = self._user.get_user_authorization_header

    def _configure_route_coverage_api(self):
        self._route_coverage = _RouteCoverageApi()
        self._configure_api_defaults(self._route_coverage)
        self.get_route_status = self._route_coverage.get_route_status
        self.get_app_route_snapshot = self._route_coverage.get_app_route_snapshot
        self.filter_routes = self._route_coverage.filter_routes

    def _configure_session_metadata_api(self):
        self._session_metadata = _SessionMetadataApi()
        self._configure_api_defaults(self._session_metadata)
        self.filter_app_session_metadata = self._session_metadata.filter_app_session_metadata



class ApplicationLibraryFilter(object):
    QuickFilterAll = 'ALL'
    QuickFilterVulnerable = 'VULNERABLE'
    QuickFilterViolation = 'VIOLATION'
    QuickFilterCustom = 'CUSTOM'
    QuickFilterNonCustom = 'NON_CUSTOM'
    QuickFilterStale = 'STALE'

    ExpandVulns = 'vulns'
    ExpandApps = 'apps'
    ExpandQuickFilters = 'quickFilters'

    def __init__(self):
        self.servers = []
        self.tags = []
        self.expand = []
        self.query = None
        self.quick_filter = ApplicationLibraryFilter.QuickFilterAll

    def get_params_as_json(self):
        return {
            'servers': ','.join(self.servers),
            'tags': ','.join(self.tags),
            'expand': ','.join(self.expand),
            'q': self.query,
            'quickFilter': self.quick_filter,
        }


from .api_support import _ApiSupport
from ..filters.library_filter import LibraryFilter

class _LibraryApi(_ApiSupport):

    def __init__(self):
        super(_LibraryApi, self).__init__()

    def get_libraries(self, org_uuid, expand=None, quick_filter='ALL'):
        path = '{org_uuid}/libraries'.format(org_uuid=org_uuid)
        return self._get(path, params={'expand': expand, 'quickFilter': quick_filter})

    def get_dotnet_library(self, org_uuid, library_hash, expand=None):
        path = '{org_uuid}/libraries/dotnet/{library_hash}'.format(org_uuid=org_uuid, library_hash=library_hash)
        return self._get(path, params={'expand': expand})

    def get_java_library(self, org_uuid, library_hash, expand=None):
        path = '{org_uuid}/libraries/java/{library_hash}'.format(org_uuid=org_uuid, library_hash=library_hash)
        return self._get(path, params={'expand': expand})

    def get_library_stats(self, org_uuid):
        path = '{org_uuid}/libraries/stats'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_library_filter_subfilters(self, org_uuid,filter_type):
        path = '{org_uuid}/libraries/filters/{filter_type}/listing'.format(org_uuid=org_uuid, filter_type=filter_type)
        return self._post(path)

    def filter_libraries(self, org_uuid, library_filter=None):
        if library_filter is None:
            library_filter = LibraryFilter()
        path = '{org_uuid}/libraries/filter'.format(org_uuid=org_uuid)
        return self._post(path, params=library_filter.get_params_as_json())

    def get_all_library_filters(self, filter_type):
        return self._get('libraries/filters/listing', params={'filterType': filter_type})

    def get_library_policy(self, org_uuid):
        path = '{org_uuid}/library/policy'.format(org_uuid=org_uuid)
        return self._get(path)




from .api_support import _ApiSupport


class _ModulesApi(_ApiSupport):

    def __init__(self):
        super(_ModulesApi, self).__init__()

    def get_application_modules(self, org_uuid, expand=None):
        path = '{org_uuid}/modules'.format(org_uuid=org_uuid)
        return self._get(path, params={'expand': expand})

    def get_application_child_modules(self, org_uuid, app_id, expand=None):
        path = '{org_uuid}/modules/{app_id}'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'expand': expand})



class TraceFilter(object):
    ExpandCard = 'card'
    ExpandEvents = 'events'
    ExpandNotes = 'notes'
    ExpandRequest = 'request'
    ExpandApplication = 'application'
    ExpandBugtracker = 'bugtracker'
    ExpandViolations = 'violations'
    ExpandSessionMetadata = 'session_metadata'

    VulnerabilityQuickFilter_All = 'all'
    VulnerabilityQuickFilter_Open = 'open'
    VulnerabilityQuickFilter_HighConfidence = 'high-confidence'
    VulnerabilityQuickFilter_Violation = 'violation'
    VulnerabilityQuickFilter_PendingReview = 'pending-review'

    TimestampFirst = 'FIRST'
    TimestampLast = 'LAST'

    def __init__(self):
        self.filter_text = None
        self.start_date = None
        self.end_date = None
        self.filter_tags = []
        self.severities = []
        self.statuses = []
        self.vuln_types = []
        self.servers = []
        self.environments = []
        self.urls = []
        self.modules = []
        self.timestamp_filter = ''
        self.expand = []
        self.quick_filter = None
        self.metadata_filters = []
        self.limit = 20
        self.offset = 0
        self.sort = '-lastTimeSeen'

    def get_params_as_json(self):
        return {
            'filterText': self.filter_text,
            'startDate': self.start_date,
            'endDate': self.end_date,
            'filterTags': ','.join(self.filter_tags),
            'severities': ','.join(self.severities),
            'status': ','.join(self.statuses),
            'vulnTypes': ','.join(self.statuses),
            'servers': ','.join(self.servers),
            'environments': ','.join(self.environments),
            'urls': ','.join(self.urls),
            'modules': ','.join(self.modules),
            'timestampFilter': self.timestamp_filter,
            'expand': ','.join(self.expand),
            'quickFilter': self.quick_filter,
            'metadataFilters': self.metadata_filters,
            'limit': self.limit,
            'offset': self.offset,
            'sort': self.sort
        }



class ServerFilter(object):
    QuickFilterAll = 'ALL'
    QuickFilterAccessEnabled = 'ACCESS_ENABLED'
    QuickFilterProtectEnabled = 'PROTECT_ENABLED'
    QuickFilterOutOfDate = 'OUT_OF_DATE'

    ExpandApplications = 'applications'
    ExpandNumberApps = 'num_apps'

    def __init__(self):
        self.application_ids = []
        self.log_levels = []
        self.tags = []
        self.agent_versions = []
        self.environments = []
        self.expand = []
        self.include_archived = False
        self.query = None
        self.quick_filter = ServerFilter.QuickFilterAll
        self.limit = 20
        self.offset = 0
        self.sort = 'lastActivity'

    def get_params_as_json(self):
        return {
            'applicationIds': ','.join(self.application_ids),
            'logLevels': ','.join(self.log_levels),
            'tags': ','.join(self.tags),
            'agentVersions': ','.join(self.agent_versions),
            'environments': ','.join(self.environments),
            'expand': ','.join(self.expand),
            'includeArchived': self.include_archived,
            'query': self.query,
            'quickFilter': self.quick_filter,
            'limit': self.limit,
            'offset': self.offset,
            'sort': self.sort
        }



from .api_support import _ApiSupport


class _UserApi(_ApiSupport):

    def __init__(self):
        super(_UserApi, self).__init__()

    def get_users(self, org_uuid, expand=None):
        path = '{org_uuid}/users'.format(org_uuid=org_uuid)
        return self._get(path, params={'expand': expand})

    def get_custom_alerts(self, org_uuid):
        path = '{org_uuid}/users/alerts/custom'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_custom_attack_alerts(self, org_uuid):
        path = '{org_uuid}/users/alerts/custom/attacks'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_custom_vulnerability_alerts(self, org_uuid):
        path = '{org_uuid}/users/alerts/custom/vulnerabilities'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_user_information(self, org_uuid, user_id, expand=None):
        path = '{org_uuid}/users/{user_id}'.format(org_uuid=org_uuid, user_id=user_id)
        return self._get(path, params={'expand': expand})

    def get_user_authorization_header(self, org_uuid, user_id):
        path = '{org_uuid}/users/{user_id}/authorization'.format(org_uuid=org_uuid, user_id=user_id)
        return self._get(path)






from .api_support import _ApiSupport


class _SessionMetadataApi(_ApiSupport):

    def __init__(self):
        super(_SessionMetadataApi, self).__init__()

    def filter_app_session_metadata(self, org_uuid, app_id):
        path = '{org_uuid}/metadata/session/{app_id}/filters'.format(org_uuid=org_uuid, app_id=app_id)
        return self._post(path)



from .api_support import _ApiSupport


class _PolicyApi(_ApiSupport):

    def __init__(self):
        super(_PolicyApi, self).__init__()

    def get_validators_and_sanitizers(self, org_uuid):
        path = '{org_uuid}/controls'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_validator_controls(self, org_uuid):
        path = '{org_uuid}/controls/validators'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_sanitizer_controls(self, org_uuid):
        path = '{org_uuid}/controls/sanitizers'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_control_suggestions(self, org_uuid):
        path = '{org_uuid}/controls/suggestion'.format(org_uuid=org_uuid)
        return self._get(path)




import requests


class _ApiSupport(object):
    _headers = None
    _base_url = None

    def __init__(self, version='ng'):
        self._version = version

    def build_url(self, path):
        return '{base_url}/{version}/{path}'.format(base_url=self._base_url,  version=self._version, path=path)

    def _get(self, path, params=None):
        return requests.get(self.build_url(path), params=params, headers=self._headers)

    def _post(self, path, data=None, params=None):
        return requests.post(self.build_url(path), json=data, headers=self._headers, params=params)

    def _put(self, path, data=None):
        return requests.put(self.build_url(path), json=data, headers=self._headers)

    def _download(self, path, params=None):
        return requests.get(self.build_url(path), params=params, headers=self._headers, stream=True)

    def _delete(self, path, data=None):
        return requests.delete(self.build_url(path), json=data, headers=self._headers)




from .api_support import _ApiSupport


class _ProfileApi(_ApiSupport):

    def __init__(self):
        super(_ProfileApi, self).__init__()

    def get_profile_info(self, expand=None):
        return self._get('profile', params=expand)

    def get_profile_organizations(self):
        return self._get('profile/organizations')

    def get_profile_default_organization(self):
        return self._get('profile/organizations/default')

    def get_org_info(self, org_uuid):
        path = 'profile/organizations/{org_uuid}'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_profile_password_policy(self):
        return self._get('profile/passwordpolicy')

    def get_profile_roles(self):
        return self._get('profile/roles')

    def set_profile_default_org(self, org_uuid):
        path = 'profile/{org_uuid}/default'.format(org_uuid=org_uuid)
        return self._put(path)





class ApplicationTraceFilter(object):
    ExpandCard = 'card'
    ExpandEvents = 'events'
    ExpandNotes = 'notes'
    ExpandRequest = 'request'
    ExpandApplication = 'application'
    ExpandServers = 'servers'

    TimestampFilterFirst = 'FIRST'
    TimestampFilterLast = 'LAST'

    def __init__(self):
        self.filter_text = None
        self.start_date = None
        self.end_date = None
        self.filter_tags = []
        self.severities = []
        self.statuses = []
        self.vuln_types = []
        self.environments = []
        self.servers = []
        self.urls = []
        self.modules = []
        self.timestamp_filter = None
        self.expand = []
        self.limit = 20
        self.offset = 0
        self.sort = '-lastTimeSeen'

    def get_params_as_json(self):
        return {
            'expand': ','.join(self.expand),
            'filterText': self.filter_text,
            'startDate': self.start_date,
            'endDate': self.end_date,
            'servers': ','.join(self.servers),
            'filterTags': ','.join(self.filter_tags),
            'severities': ','.join(self.severities),
            'environments': ','.join(self.environments),
            'status': ','.join(self.statuses),
            'vulnTypes': ','.join(self.vuln_types),
            'modules': self.modules,
            'urls': self.urls,
            'limit': self.limit,
            'offset': self.offset,
            'sort': self.sort
        }


from .api_support import _ApiSupport
from ..filters.server_filter import ServerFilter
from ..filters.server_library_filter import ServerLibraryFilter
from ..filters.server_trace_filter import ServerTraceFilter


class _ServerApi(_ApiSupport):

    def __init__(self):
        super(_ServerApi, self).__init__()

    def get_servers(self, org_uuid, include_archived=False, expand=None, query=None):
        path = '{org_uuid}/servers'.format(org_uuid=org_uuid)
        return self._get(path, params={'q': query, 'includeArchived': include_archived, 'expand': expand})

    def get_active_servers(self, org_uuid):
        path = '{org_uuid}/servers/active'.format(org_uuid=org_uuid)
        return self._get(path)

    def filter_servers(self, org_uuid, server_filter=None):
        if server_filter is None:
            server_filter = ServerFilter()
        path = '{org_uuid}/servers/filter'.format(org_uuid=org_uuid)
        return self._post(path, params=server_filter.get_params_as_json())

    def get_server_filters(self, org_uuid):
        path = '{org_uuid}/servers/filters/listing'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_server_filter_subfilters(self, org_uuid, filter_type):
        path = '{org_uuid}/servers/filters/{filter_type}/listing'.format(org_uuid=org_uuid, filter_type=filter_type)
        return self._post(path)

    def get_server_modes(self, org_uuid):
        path = '{org_uuid}/servers/modes'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_server_details(self, org_uuid, server_id, expand=None):
        path = '{org_uuid}/servers/{server_id}'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'expand': expand})

    def get_server_activity(self, org_uuid, server_id, start_date=None, end_date=None):
        path = '{org_uuid}/servers/{server_id}/activity'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'startDate': start_date, 'endDate': end_date})

    def get_server_agent_activity(self, org_uuid, server_id, interval='WEEK'):
        path = '{org_uuid}/servers/{server_id}/activity/interval'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'interval': interval})

    def get_server_app_traces(self, org_uuid, server_id, orphans=False):
        path = '{org_uuid}/servers/{server_id}/apptraces'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'orphans': orphans})

    def get_server_attack_status(self, org_uuid, server_id, include_merged=True):
        path = '{org_uuid}/servers/{server_id}/breakdown/attack/status'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'includeMerged': include_merged})

    def get_server_attack_types(self, org_uuid, server_id, include_merged=True):
        path = '{org_uuid}/servers/{server_id}/breakdown/attack/type'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'includeMerged': include_merged})

    def get_server_trace_breakdown(self, org_uuid, server_id, include_merged=True):
        path = '{org_uuid}/servers/{server_id}/breakdown/trace/rule'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'includeMerged': include_merged})

    def get_server_trace_severity_breakdown(self, org_uuid, server_id, include_merged=True):
        path = '{org_uuid}/servers/{server_id}/breakdown/trace/severity'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'includeMerged': include_merged})

    def get_server_trace_status_breakdown(self, org_uuid, server_id, include_merged=True):
        path = '{org_uuid}/servers/{server_id}/breakdown/trace/status'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'includeMerged': include_merged})

    def get_server_libraries_breakdown(self, org_uuid, server_id, include_merged=True, include_archived=False):
        path = '{org_uuid}/servers/{server_id}/libraries/breakdown'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'includeMerged': include_merged, 'includeArchived': include_archived})

    def update_server_name(self, org_uuid, server_id, new_name):
        path = '{org_uuid}/servers/{server_id}/name'.format(org_uuid=org_uuid, server_id=server_id)
        return self._put(path, data={'name': new_name})

    def get_server_properties(self, org_uuid, server_id):
        path = '{org_uuid}/servers/{server_id}/properties'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path)

    def get_server_vuln_and_attack_urls(self, org_uuid, server_id, interval='WEEK'):
        path = '{org_uuid}/servers/{server_id}/url'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'interval': interval})

    def get_server_vuln_urls(self, org_uuid, server_id, interval='WEEK'):
        path = '{org_uuid}/servers/{server_id}/url/vuln'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'interval': interval})

    def get_server_attack_urls(self, org_uuid, server_id, interval='WEEK'):
        path = '{org_uuid}/servers/{server_id}/url/attack'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'interval': interval})

    def get_server_libraries(self, org_uuid, server_id, expand=None, quick_filter='ALL'):
        path = '{org_uuid}/servers/{server_id}/libraries'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'expand': expand, 'quickFilter': quick_filter})

    def get_server_libraries_subfilters(self, org_uuid, server_id, libraries_filter_type):
        path = '{org_uuid}/servers/{server_id}/libraries/filters/{libraries_filter_type}/listing'.format(org_uuid=org_uuid, server_id=server_id, libraries_filter_type=libraries_filter_type)
        return self._get(path)

    def get_server_libraries_stats(self, org_uuid, server_id, include_merged=True):
        path = '{org_uuid}/servers/{server_id}/libraries/stats'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params={'includeMerged': include_merged})

    def filter_server_libraries(self, org_uuid, server_id, server_library_filter=None):
        if server_library_filter is None:
            server_library_filter = ServerLibraryFilter()
        path = '{org_uuid}/servers/{server_id}/libraries/filter'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params=server_library_filter.get_params_as_json())

    def get_server_trace_subfilters(self, org_uuid, server_id, trace_filter_type):
        path = '{org_uuid}/servertraces/{server_id}/filter/{trace_filter_type}/listing'.format(org_uuid=org_uuid, server_id=server_id, trace_filter_type=trace_filter_type)
        return self._get(path)

    def get_server_trace_details(self, org_uuid, server_id, trace_uuid, expand=None):
        path = '{org_uuid}/servertraces/{server_id}/filter/{trace_uuid}'.format(org_uuid=org_uuid, server_id=server_id, trace_uuid=trace_uuid)
        return self._get(path, params={'expand': expand})

    def filter_server_traces(self, org_uuid, server_id, server_trace_filter=None):
        if server_trace_filter is None:
            server_trace_filter = ServerTraceFilter()
        path = '{org_uuid}/servertraces/{server_id}/filter'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path, params=server_trace_filter.get_params_as_json())

    def delete_server_traces(self, org_uuid, server_id, traces=[]):
        if len(traces) < 1:
            raise BaseException('Trace list cannot be empty')
        path = '{org_uuid}/servertraces/{server_id}'.format(org_uuid=org_uuid, server_id=server_id)
        return self._delete(path, data={'traces': traces})

    def get_server_policy_violations(self, org_uuid, server_id):
        path = '{org_uuid}/servertraces/{server_id}/policy/violations'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path)

    def delete_server_trace(self, org_uuid, server_id, trace_uuid):
        path = '{org_uuid}/servertraces/{server_id}/trace/{trace_uuid}'.format(org_uuid=org_uuid, server_id=server_id, trace_uuid=trace_uuid)
        return self._get(path)

    def get_server_trace_vulnerability(self, org_uuid, server_id, trace_uuid):
        path = '{org_uuid}/servertraces/{server_id}/{trace_uuid}/visible'.format(org_uuid=org_uuid, server_id=server_id, trace_uuid=trace_uuid)
        return self._get(path);


from .api_support import _ApiSupport


class _HistoryApi(_ApiSupport):

    def __init__(self):
        super(_HistoryApi, self).__init__()

    def get_organization_score_history(self, org_uuid, limit=12):
        path = '{org_uuid}/history/scores'.format(org_uuid=org_uuid)
        return self._get(path, params={'limit': limit})

    def get_organization_score_history_interval(self, org_uuid, interval='WEEK', include_defense=False):
        defense_string = '/defense' if include_defense else ''
        path = '{org_uuid}/history/scores/interval{defense_string}'.format(org_uuid=org_uuid, defense_string=defense_string)
        return self._get(path, params={'interval': interval})


from .api_support import _ApiSupport


class _OrganizationApi(_ApiSupport):

    def __init__(self):
        super(_OrganizationApi, self).__init__()

    def search(self, org_uuid, query_string):
        path = '{org_uuid}/search'.format(org_uuid=org_uuid)
        return self._get(path, params={'q': query_string})

    def get_organization_info(self, org_uuid, expand=None):
        path = '{org_uuid}/organizations'.format(org_uuid=org_uuid)
        return self._get(path, params={'expand': expand})

    def get_organization_administrators(self, org_uuid):
        path = '{org_uuid}/organizations/administrators'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_organization_application_roles(self, org_uuid):
        path = '{org_uuid}/organizations/application/roles'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_organization_library_scoring(self, org_uuid):
        path = '{org_uuid}/organizations/scoring/libraries'.format(org_uuid=org_uuid)
        return self._get(path)

    def put_organization_library_scoring(self, org_uuid, fail_libraries_policy=True, scoring_type=None):
        path = '{org_uuid}/organizations/scoring/libraries'.format(org_uuid=org_uuid)
        return self._put(path, data={'fail_libraries_policy': fail_libraries_policy, 'type': scoring_type})

    def get_organization_servers_needing_restart(self, org_uuid, language):
        path = '{org_uuid}/organizations/servers/restart/{language}'.format(org_uuid=org_uuid, language=language)
        return self._get(path)

    def get_organization_application_stats(self, org_uuid, interval='WEEK', expand=None):
        path = '{org_uuid}/organizations/stats/application'.format(org_uuid=org_uuid)
        return self._get(path, params={'interval': interval, 'expand': expand})

    def get_organization_library_stats(self, org_uuid, interval='WEEK', expand=None):
        path = '{org_uuid}/organizations/stats/library'.format(org_uuid=org_uuid)
        return self._get(path, params={'interval': interval, 'expand': expand})

    def get_organization_server_stats(self, org_uuid, interval='WEEK', expand=None):
        path = '{org_uuid}/organizations/stats/server'.format(org_uuid=org_uuid)
        return self._get(path, params={'interval': interval, 'expand': expand})

    def get_organization_trace_stats(self, org_uuid, interval='WEEK'):
        path = '{org_uuid}/organizations/stats/trace'.format(org_uuid=org_uuid)
        return self._get(path, params={'interval': interval})

    def get_organization_server_settings(self, org_uuid):
        path = '{org_uuid}/server/settings'.format(org_uuid=org_uuid)
        return self._get(path)






from .api_support import _ApiSupport


class _RouteCoverageApi(_ApiSupport):

    def __init__(self):
        super(_RouteCoverageApi, self).__init__()

    def get_route_status(self, org_uuid, app_id):
        path = '{org_uuid}/applications/{app_id}/route/status'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path)

    def get_app_route_snapshot(self, org_uuid, app_id, start_date=None, end_date=None):
        """

        :param org_uuid: organization id
        :param app_id: application id
        :param start_date: int version of timestamp; Ex Jan 1, 2020 -> 1577854800
        :param end_date: int version of timestamp; Ex Jan 1, 2020 -> 1577854800
        :return: json of app route snapshot
        """
        path = '{org_uuid}/applications/{app_id}/route/snapshots'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path, params={'startDate': start_date, 'endDate': end_date})

    def filter_routes(self, org_uuid, app_id, metadata=None, sessionID='', expand=None):
        """
        Metadata is a list of dicts containing a label and a list of values

        Example:

        metadata = [{'label': 'test_label', 'values': ['test_value_1']}]
        """
        if metadata is None:
            metadata = []
        path = '{org_uuid}/applications/{app_id}/route/filter'.format(org_uuid=org_uuid, app_id=app_id)
        return self._post(path, data={'metadata': metadata, 'sessionID': sessionID}, params={'expand': expand})



from .api_support import _ApiSupport


class _WebhookApi(_ApiSupport):

    def __init__(self):
        super(_WebhookApi, self).__init__()

    def get_webhooks(self, org_uuid, expand=None):
        path = '{org_uuid}/webhooks'.format(org_uuid=org_uuid)
        return self._get(path, params={'expand': expand})

    def get_webhook(self, org_uuid, webhook_id, expand=None):
        path = '{org_uuid}/webhooks/{webhook_id}'.format(org_uuid=org_uuid, webhook_id=webhook_id)
        return self._get(path, params={'expand': expand})



from .api_support import _ApiSupport


class _TagsApi(_ApiSupport):

    def __init__(self):
        super(_TagsApi, self).__init__()

    def get_application_tags(self, org_uuid, app_id):
        path = '{org_uuid}/tags/application/list/{app_id}'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path)

    def get_all_library_tags(self, org_uuid):
        path = '{org_uuid}/tags/libraries/list'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_all_application_tags(self, org_uuid):
        path = '{org_uuid}/tags/applications/list'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_application_library_tags(self, org_uuid, app_id):
        path = '{org_uuid}/tags/libraries/{app_id}/list'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path)

    def get_library_tag_list(self, org_uuid, library_hash):
        path = '{org_uuid}/tags/library/list/{library_hash}'.format(org_uuid=org_uuid, library_hash=library_hash)
        return self._get(path)

    def get_server_tag_list(self, org_uuid, server_id):
        path = '{org_uuid}/tags/server/list/{server_id}'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path)

    def get_all_server_tags(self, org_uuid):
        path = '{org_uuid}/tags/servers/list'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_all_trace_tags(self, org_uuid):
        path = '{org_uuid}/tags/traces'.format(org_uuid=org_uuid)
        return self._get(path)

    def get_all_trace_tags_for_application(self, org_uuid, app_id):
        path = '{org_uuid}/tags/traces/application/{app_id}'.format(org_uuid=org_uuid, app_id=app_id)
        return self._get(path)

    def get_all_trace_tags_for_servers(self, org_uuid, server_id):
        path = '{org_uuid}/tags/traces/server/{server_id}'.format(org_uuid=org_uuid, server_id=server_id)
        return self._get(path)

    def get_all_tags_for_trace(self, org_uuid, trace_id):
        path = '{org_uuid}/tags/traces/trace/{trace_id}'.format(org_uuid=org_uuid, trace_id=trace_id)
        return self._get(path)

    def tag_application(self, org_uuid, app_id, tag_name):
        path = '{org_uuid}/tags/applications'.format(org_uuid=org_uuid)
        return self._put(path, data={"applications_id": [app_id], 'links': [], 'tags': [tag_name]})

    def tag_library(self, org_uuid, library, tag_name):
        path = '{org_uuid}/tags/libraries'.format(org_uuid=org_uuid)
        return self._put(path, data={'libraries': [library], 'links': [], 'tags': [tag_name]})

    def tag_server(self, org_uuid, server_id, tag_name):
        path = '{org_uuid}/tags/servers'.format(org_uuid=org_uuid)
        return self._put(path, data={'servers_id': [server_id], 'links': [], 'tags': [tag_name]})

    def tag_trace(self, org_uuid, trace_id, tag_name):
        path = '{org_uuid}/tags/traces'.format(org_uuid=org_uuid)
        return self._put(path, data={'traces_id': [trace_id], 'links': [], 'tags': [tag_name]})

    def delete_tag_from_application(self, org_uuid, app_id, tag):
        path = '{org_uuid}/tags/application/{app_id}'.format(org_uuid=org_uuid, app_id=app_id)
        return self._delete(path, data={'tag': tag})

    def delete_tag_from_trace(self, org_uuid, trace_id, tag):
        path = '{org_uuid}/tags/trace/{trace_id}'.format(org_uuid=org_uuid, trace_id=trace_id)
        return self._delete(path, data={'tag': tag})

    def delete_tag_from_server(self, org_uuid, server_id, tag):
        path = '{org_uuid}/tags/server/{server_id}'.format(org_uuid=org_uuid, server_id=server_id)
        return self._delete(path, data={'tag': tag})

    def delete_tag_from_library(self, org_uuid, library_hash, tag):
        path = '{org_uuid}/tags/library/{library_hash}'.format(org_uuid=org_uuid, library_hash=library_hash)
        return self._delete(path, data={'tag':tag})


class TraceTimeToRemediateFilter(object):

    def __init__(self):
        self.apps = []
        self.importance = []
        self.environments = []
        self.servers = []
        self.tags = []
        self.interval = 'WEEK'
        self.expand = []

    def get_params_as_json(self):
        return {
            'apps':  ','.join(self.apps),
            'importance': ','.join(self.importance),
            'environments': ','.join(self.environments),
            'servers': ','.join(self.servers),
            'tags': ','.join(self.tags),
            'expand': ','.join(self.expand),
            'interval': self.interval
        }